@@ -1,15 +1,19 @@
 from __future__ import annotations
 
 import logging
-from datetime import datetime
+from datetime import datetime, UTC
 from enum import StrEnum, auto
 from typing import Literal
 
 import openpectus.aggregator.models as Mdl
+import webpush
 from pydantic import BaseModel, ConfigDict
 from pydantic.json_schema import SkipJsonSchema
 
 SystemStateEnum = Mdl.SystemStateEnum
+NotificationScope = Mdl.NotificationScope
+NotificationTopic = Mdl.NotificationTopic
+WebPushSubscription = webpush.WebPushSubscription
 
 
 class Dto(BaseModel):
@@ -119,12 +123,18 @@ class ProcessUnit(Dto):
     current_user_role: UserRole
     uod_author_name: str | SkipJsonSchema[None] = None
     uod_author_email: str | SkipJsonSchema[None] = None
+
     # users: list[User] ?
 
     def __str__(self) -> str:
         return f'{self.__class__.__name__}(id="{self.id}", name="{self.name}", state={self.state})'
 
 
+class ProcessUnitAllProcessValues(Dto):
+    process_unit: ProcessUnit
+    process_values: list[ProcessValue]
+
+
 class ProcessValueType(StrEnum):
     STRING = auto()
     FLOAT = auto()
@@ -288,8 +298,8 @@ class RunLogLine(Dto):
                 source=CommandSource.METHOD,
                 value=None
             ),
-            start=datetime.fromtimestamp(model.start),
-            end=None if model.end is None else datetime.fromtimestamp(model.end),
+            start=datetime.fromtimestamp(model.start, UTC),
+            end=None if model.end is None else datetime.fromtimestamp(model.end, UTC),
             progress=model.progress,
             start_values=list(map(ProcessValue.create, model.start_values)),
             end_values=list(map(ProcessValue.create, model.end_values)),
@@ -341,9 +351,11 @@ class Method(Dto):
             last_author=method.last_author
         )
 
+
 class MethodVersion(Dto):
     version: int
 
+
 class MethodState(Dto):
     started_line_ids: list[str]
     executed_line_ids: list[str]
@@ -363,6 +375,7 @@ class MethodState(Dto):
                            executed_line_ids=[_id for _id in method_state.executed_line_ids],
                            injected_line_ids=[_id for _id in method_state.injected_line_ids])
 
+
 class MethodAndState(Dto):
     method: Method
     state: MethodState
@@ -456,6 +469,11 @@ class PlotLog(Dto):
     entries: dict[str, PlotLogEntry]
 
 
+class Contributor(Dto):
+    id: str | SkipJsonSchema[None] = None
+    name: str
+
+
 class RecentRun(Dto):
     """ Represents a historical run of a process unit. """
     engine_id: str
@@ -470,7 +488,7 @@ class RecentRun(Dto):
     engine_hardware_str: str
     aggregator_computer_name: str
     aggregator_version: str
-    contributors: list[str] = []
+    contributors: list[Contributor] = []
 
     def __str__(self) -> str:
         return f'{self.__class__.__name__}(engine_id="{self.engine_id}", run_id="{self.run_id}")'
@@ -518,6 +536,7 @@ class AggregatedErrorLogEntry(Dto):
             occurrences=model.occurrences
         )
 
+
 class AggregatedErrorLog(Dto):
     entries: list[AggregatedErrorLogEntry]
 
@@ -545,10 +564,12 @@ class TagDefinition(Dto):
     unit: str | None = None
     # possibly value_type:
 
+
 class CommandDefinition(Dto):
     name: str
     validator: str | None = None
 
+
 class UodDefinition(Dto):
     # name: str
     # filename: str
@@ -563,3 +584,13 @@ class UodDefinition(Dto):
             system_commands=[CommandDefinition(name=c.name, validator=c.validator) for c in model.system_commands],
             tags=[TagDefinition(name=t.name, unit=t.unit) for t in model.tags]
         )
+
+
+class WebPushConfig(Dto):
+    app_server_key: str | SkipJsonSchema[None] = None
+
+
+class WebPushNotificationPreferences(Dto):
+    scope: NotificationScope
+    topics: list[NotificationTopic]
+    process_units: list[str]
