@@ -1,4 +1,5 @@
 import logging
+import asyncio
 
 import openpectus.aggregator.deps as agg_deps
 import openpectus.aggregator.models as Mdl
@@ -138,9 +139,34 @@ def get_all_process_values(
     return process_values
 
 
+@router.get("/process_units/all_process_values", response_model_exclude_none=True)
+def get_all_process_values_of_all_available_engines(
+        user_roles: UserRolesValue,
+        response: Response,
+        agg: Aggregator = Depends(agg_deps.get_aggregator)
+) -> list[Dto.ProcessUnitAllProcessValues]:
+    """
+    Returns all process value for all online process units (engines)
+    that the current user has access to.
+    """
+    response.headers["Cache-Control"] = "no-store"
+    process_units: list[Dto.ProcessUnitAllProcessValues] = []
+    all_engine_data = agg.get_all_registered_engine_data()
+    for engine_data in all_engine_data:
+        if not has_access(engine_data, user_roles):
+            continue
+        process_unit = map_pu(engine_data)
+        process_units.append(Dto.ProcessUnitAllProcessValues(
+            process_unit=process_unit,
+            process_values=[Dto.ProcessValue.create(tag_value) for tag_value in engine_data.tags_info.map.values()],
+        ))
+    return process_units
+
+
 @router.post("/process_unit/{unit_id}/execute_command", response_model_exclude_none=True)
 async def execute_command(
         user_name: UserNameValue,
+        user_id: UserIdValue,
         user_roles: UserRolesValue,
         unit_id: str,
         command: Dto.ExecutableCommand,
@@ -162,7 +188,10 @@ async def execute_command(
 
     # for now, all users issuing a command become contributors. may nee to filter that somehow
     # and when wo we clear the contributors?
-    engine_data.contributors.add(user_name)
+    contributor = Mdl.Contributor(id=user_id, name=user_name)
+    if contributor not in engine_data.contributors:
+            agg.from_frontend.publish_new_contributor_notification(unit_id, contributor)
+    engine_data.contributors.add(contributor)
     return Dto.ServerSuccessResponse()
 
 
@@ -223,6 +252,7 @@ def get_method(
 @router.post('/process_unit/{unit_id}/method', response_model_exclude_none=True)
 async def save_method(
         user_name: UserNameValue,
+        user_id: UserIdValue,
         user_roles: UserRolesValue,
         unit_id: str,
         method_dto: Dto.Method,
@@ -233,7 +263,7 @@ async def save_method(
         version=method_dto.version,
         last_author=user_name
     )
-    new_version = await agg.from_frontend.method_saved(engine_id=unit_id, method=method_mdl, user_name=user_name)
+    new_version = await agg.from_frontend.method_saved(engine_id=unit_id, method=method_mdl, user=Mdl.Contributor(id=user_id, name=user_name))
     return Dto.MethodVersion(version=new_version)
 
 
@@ -291,12 +321,13 @@ def get_error_log(
 @router.post('/process_unit/{unit_id}/run_log/force_line/{line_id}', response_model_exclude_none=True)
 async def force_run_log_line(
         user_name: UserNameValue,
+        user_id: UserIdValue,
         user_roles: UserRolesValue,
         unit_id: str,
         line_id: str,
         agg: Aggregator = Depends(agg_deps.get_aggregator)):
     _ = get_registered_engine_data_or_fail(unit_id, user_roles, agg)
-    if not await agg.from_frontend.request_force(engine_id=unit_id, line_id=line_id, user_name=user_name):
+    if not await agg.from_frontend.request_force(engine_id=unit_id, line_id=line_id, user=Mdl.Contributor(id=user_id, name=user_name)):
         return Dto.ServerErrorResponse(message="Force request failed")
     return Dto.ServerSuccessResponse(message="Force successfully requested")
 
@@ -304,12 +335,13 @@ async def force_run_log_line(
 @router.post('/process_unit/{unit_id}/run_log/cancel_line/{line_id}', response_model_exclude_none=True)
 async def cancel_run_log_line(
         user_name: UserNameValue,
+        user_id: UserIdValue,
         user_roles: UserRolesValue,
         unit_id: str,
         line_id: str,
         agg: Aggregator = Depends(agg_deps.get_aggregator)):
     _ = get_registered_engine_data_or_fail(unit_id, user_roles, agg)
-    if not await agg.from_frontend.request_cancel(engine_id=unit_id, line_id=line_id, user_name=user_name):
+    if not await agg.from_frontend.request_cancel(engine_id=unit_id, line_id=line_id, user=Mdl.Contributor(id=user_id, name=user_name)):
         return Dto.ServerErrorResponse(message="Cancel request failed")
     return Dto.ServerSuccessResponse(message="Cancel successfully requested")
 
