@@ -3,7 +3,7 @@ import itertools
 import logging
 import uuid
 from queue import Empty, Queue
-from typing import Iterable, Set
+from typing import Iterable, Literal, Set
 from uuid import UUID
 from openpectus.engine.internal_commands import InternalCommandsRegistry
 from openpectus.engine.hardware import HardwareLayerException, RegisterDirection
@@ -144,12 +144,8 @@ class Engine(InterpreterContext):
         self._last_error : Exception | None = None
         """ Cause of error_state"""
 
-        self._method_manager: MethodManager = MethodManager(uod.get_command_names())
-        """ The model handling changes to method code """
-        self._pending_merge_method: Mdl.Method | None = None
-
-        self._interpreter: PInterpreter = PInterpreter(self._method_manager.program, self)
-        """ The interpreter executing the current method. """
+        self._method_manager: MethodManager = MethodManager(uod.get_command_names(), self)
+        """ The model handling changes to method code and interpreter running it """
 
         self._cancel_command_exec_ids: set[UUID] = set()
 
@@ -182,18 +178,18 @@ class Engine(InterpreterContext):
     def base_unit_provider(self) -> BaseUnitProvider:
         return self.uod.base_unit_provider
 
+    @property
+    def method_manager(self) -> MethodManager:
+        return self._method_manager
+
     @property
     def interpreter(self) -> PInterpreter:
-        assert self._interpreter is not None
-        return self._interpreter
+        return self.method_manager.interpreter
 
     @property
     def runtimeinfo(self) -> RuntimeInfo:
         return self.interpreter.runtimeinfo
 
-    @property
-    def method_manager(self) -> MethodManager:
-        return self._method_manager
 
     def cleanup(self):
         self.emitter.emit_on_engine_shutdown()
@@ -253,19 +249,6 @@ class Engine(InterpreterContext):
             for tag in self._system_tags.tags.values():
                 tag.tick_time = tick_time
 
-        # edit phase
-        if self._pending_merge_method is not None:
-            logger.debug("Applying scheduled method merge")
-            try:
-                self._method_manager.merge_method(self._pending_merge_method)
-                self._interpreter.update_method_and_ffw(self._method_manager.program)
-                logger.debug("Method edit complete")
-            except Exception as ex:
-                logger.error("Method edit failed", exc_info=True)
-                self.set_error_state(ex)
-            finally:
-                self._pending_merge_method = None
-
         self.uod.hwl.tick()
 
         # read phase, error_state on HardwareLayerException
@@ -279,7 +262,7 @@ class Engine(InterpreterContext):
                 not self._runstate_stopping:
             try:
                 # run one tick of interpretation, i.e. one instruction
-                self._interpreter.tick(tick_time, self._tick_number)
+                self.interpreter.tick(tick_time, self._tick_number)
             except InterpretationInternalError as ex:
                 logger.fatal("A serious internal interpreter error occured. The method should be stopped. If it is resumed, \
                              additional errors may occur.", exc_info=True)
@@ -447,7 +430,7 @@ class Engine(InterpreterContext):
             if args is not None:
                 try:
                     command.validate_arguments(args)
-                    logger.debug(f"Initialized command {cmd_request.name} with arguments {args}")
+                    logger.debug(f"Initialized command {cmd_request.name} with arguments '{args}'")
                 except Exception:
                     raise EngineError(
                         f"Failed to initialize arguments '{args}' for command '{cmd_request.name}'",
@@ -482,8 +465,16 @@ class Engine(InterpreterContext):
         self._system_tags[SystemTagName.RUN_ID].set_value(None, self._tick_time)
 
     def _stop_interpreter(self):
-        self._interpreter.stop()
-        self._interpreter = PInterpreter(self.method_manager.program, self)
+        # called from Stop and Restart
+
+        # does not seem necessary if calling reset_interpreter() below but it is
+        self.interpreter.stop()
+
+        # TODO: This hack reproduces the behavior of setting new interpreter
+        # a few tests related to start/stop/restart depend on this - but it is not
+        # clear whether this is a real requirement or not. In fact it may make more
+        # sense to wait until start
+        self.method_manager.reset_interpreter()
 
     def _cancel_uod_commands(self):
         logger.debug("Cancelling uod commands")
@@ -638,6 +629,7 @@ class Engine(InterpreterContext):
                     cmd_request.command_exec_id,
                     self._tick_time, self._tick_number,
                     self.tags_as_readonly())
+                record.node.completed = True
                 cmds_done.add(cmd_request)
                 uod_command.finalize()
                 logger.debug(f"Command {cmd_request.name} finalized")
@@ -665,15 +657,12 @@ class Engine(InterpreterContext):
 
     def _apply_safe_state(self) -> TagValueCollection:
         current_values: list[TagValue] = []
-
-        # TODO we should probably only consider uod tags here. would system tags ever have a safe value?
-        for t in self._iter_all_tags():
-            if t.direction == TagDirection.Output:
-                safe_value = t.safe_value
-                if not isinstance(safe_value, Unset):
-                    current_values.append(t.as_readonly())
-                    t.set_value(safe_value, self._tick_time)
-
+        hwl = self.uod.hwl
+        registers = [r for r in hwl.registers.values() if RegisterDirection.Write in r.direction and "safe_value" in r._options]
+        for r in registers:
+            tag = self.uod.tags[r.name]
+            current_values.append(tag.as_readonly())
+            tag.set_value(r._options["safe_value"], self._tick_time)
         return TagValueCollection(current_values)
 
     def _apply_state(self, state: TagValueCollection):
@@ -686,7 +675,7 @@ class Engine(InterpreterContext):
         """ Collect tag updates from all tags, even unmodified tags """
         for tag in self._iter_all_tags():
             if tag.tick_time is None:
-                logger.warning(f'Setting a tick time on {tag.name} tag missing it in notify_initial_tags()')
+                logger.warning(f'Setting a tick time on {tag.name} tag missing it in notify_all_tags()')
                 tag.tick_time = self._tick_time
             self.tag_updates.put(tag)
 
@@ -696,6 +685,11 @@ class Engine(InterpreterContext):
         for tag_name in self._system_listener.changes:
             tag = self._system_tags[tag_name]
             self.tag_updates.put(tag)
+            if tag_name == SystemTagName.CONNECTION_STATUS:
+                status = self._system_tags[SystemTagName.CONNECTION_STATUS].get_value()
+                assert isinstance(status, str)
+                assert status == "Disconnected" or status == "Connected"
+                self._emitter.emit_on_connection_status_change(status)
         self._system_listener.clear_changes()
 
         for tag_name in self._uod_listener.changes:
@@ -709,6 +703,7 @@ class Engine(InterpreterContext):
         self._system_tags[SystemTagName.SYSTEM_STATE].set_value(SystemStateEnum.Paused, self._tick_time)
         self._last_error = exception
         self._runstate_paused = True
+        self._emitter.emit_on_method_error(exception)
 
     def has_error_state(self) -> bool:
         method_status = self._system_tags[SystemTagName.METHOD_STATUS]
@@ -775,57 +770,38 @@ class Engine(InterpreterContext):
             raise
 
     # code manipulation api
-    def set_method(self, method: Mdl.Method):
-        """ Set new method. This will replace the current method.
-
-        We generally expect that nothing important is going on when performing a method edit. This means
-        that interpreter is in one of these states:
-            - Current instruction is awating a threshold, preferably with at least several seconds before
-              the instruction starts
-            - Current instruction is a Wait
-            - Current instruction is a Pause
-            - Method is exhausted
-            - Engine paused
-            - ??
-            In the waiting cases (threshold, Wait and Pause), the waiting time is reset - this is probably ok?
-
-        How should this be enforced?
-        - Wait for an active instruction to complete before performing the edit? Probably. This does not mean
-          wait for a possible command, just wait to engine phase edit
-        - Fail unless one of the conditions above is true? This is difficult:
-            - This likely requires semantic changes so no two instructions are active in the same tick.
-              - A case is that currently two adjecent Mark instruction are active in the same tick
-                and we have tests that depend on this - but this can be changed and probably should. It would
-                certainly make tests easier to write and understand and more robust since they don't have to depend
-                on tick counting.
-        - Should the edit be performed as a command? We should always perform the edit in a "pause" so it won't interfere
-          with anything...
-
-        If a method is already running, it will
-        - Take node of the current instruction
-        - Check that no code changes exist in already executed code. If so, MethodEditError is raised
-        - Serialize the AST of the running method
-        - Parse the updated method, replacing the AST with the new AST
-        - Apply persisted data to new AST
-        - Patch all node references to point to the updated nodes:
-            - In runtime records
-            - In interpreter state (activate records)
-        - Start executing new program in fast-forward mode to skip directly to the paused instruction
-        """
+    def set_method(self, method: Mdl.Method) -> Literal["merge_method", "set_method"]:
+        """ Set new method. This will replace the current method, either by merging in changes in case the method is already
+        running or just setting the method otherwise. """
 
         try:
-            if self._runstate_started:
-                if self._pending_merge_method is not None:
-                    raise MethodEditError("An edit is already in progress")
-                # signal to apply updated method on next tick
-                self._pending_merge_method = method
+            if self._runstate_started and self.method_manager.program_is_started:
+                logger.info(f"Method changed while running. Current revision {self.method_manager.program.revision}")
+                try:
+                    self._method_manager.merge_method(method)
+                    logger.info(f"Method merged successfully. Revision is now {self.method_manager.program.revision}")
+                    return "merge_method"
+                except Exception:
+                    logger.error("Error merging method")
+                    raise
             else:
-                self._method_manager.set_method(method)
-                self._interpreter = PInterpreter(self._method_manager.program, self)
-                logger.debug(f"New method set with {len(method.lines)} lines")
+                logger.info("Setting new method")
+                try:
+                    self._method_manager.set_method(method)
+                    logger.info(f"Method set successfully. Revision is {self.method_manager.program.revision}")
+                    return "set_method"
+                except Exception:
+                    logger.error("Error setting method")
+                    raise
 
+        except MethodEditError:
+            # skip setting error_state(?) tests expects this and fails on error_state
+            # depends on how we will handle these errors on the client side
+            raise
         except Exception as ex:
             logger.error("Failed to set method", exc_info=True)
+            logger.error(f"Current method content:\n\n{self.method_manager._method}\n")
+            logger.error(f"New method content:\n\n{method}\n")
             self.set_error_state(ex)
             raise
 
