@@ -184,6 +184,8 @@ class StopEngineCommand(InternalEngineCommand):
                     break
                 yield
             e._finalize_uod_commands()
+            e._apply_safe_state()
+            e.write_process_image()
 
             logger.debug("All uod commands have completed execution. Stop will now complete.")
             e._runstate_started = False
