@@ -2,10 +2,8 @@
 Defines the standard imports for uod modules.
 """
 from openpectus.engine.hardware import HardwareLayerBase, Register, RegisterDirection
-from openpectus.lang.exec.uod import (
-    UnitOperationDefinitionBase, UodCommand, UodBuilder,
-    RegexNumber, RegexText, RegexCategorical,
-)
+from openpectus.lang.exec.regex import RegexNumber, RegexText, RegexCategorical
+from openpectus.lang.exec.uod import UnitOperationDefinitionBase, UodCommand, UodBuilder
 from openpectus.protocol.models import PlotConfiguration, SubPlot, PlotAxis, PlotColorRegion
 import openpectus.lang.exec.tags_impl as tags
 import openpectus.lang.exec.units as units
