@@ -71,6 +71,7 @@ class RuntimeInfo:
                 runlog.items.extend(items)
             except Exception:
                 logger.error(f"Failed to create runlog item for record {r}", exc_info=True)
+                raise
         runlog.items.sort(key=lambda item: item.start)
         return runlog
 
@@ -80,10 +81,10 @@ class RuntimeInfo:
                 return str(r.node.id)
 
     def _get_record_runlog_items(self, r: RuntimeRecord) -> list[RunLogItem]:  # noqa C901
-        if isinstance(r.node, (p.ProgramNode, p.BlankNode, p.CommentNode)):
+        if isinstance(r.node, (p.ProgramNode, p.BlankNode, p.CommentNode, p.InjectedNode)):
             return []
         if r.name is None:
-            if isinstance(r.node, (p.InjectedNode, p.ErrorInstructionNode)):
+            if isinstance(r.node, (p.ErrorInstructionNode)):
                 return []
             node_name = str(r.node) if r.node is not None else "node is None"
             logger.error(f"Runtime record has empty name. node: {node_name}. Fix this error or add a rule exception.")
@@ -151,7 +152,7 @@ class RuntimeInfo:
                     item.id = str(state.command_exec_id)
                     command = state.command
                 elif state.state_name == RuntimeRecordStateEnum.AwaitingThreshold:
-                    assert item is not None
+                    assert item is not None, f"Item for record {r} was unexpectedly None in state {state.state_name}. States: {invocation_states}"
                     item.state = RunLogItemState.AwaitingThreshold
 
                 if not is_conclusive_state and item is not None:
@@ -167,7 +168,7 @@ class RuntimeInfo:
                 if is_conclusive_state:
                     assert item is not None
                     item.end = state.state_time
-                    item.end_values = state.values or TagValueCollection. empty()
+                    item.end_values = state.values or TagValueCollection.empty()
                     item.cancellable = False
                     item.forcible = False
 
@@ -274,14 +275,16 @@ class RuntimeInfo:
                         return state.command, record
 
     def get_last_node_record_or_none(self, node: p.Node) -> RuntimeRecord | None:
+        """ Get last record by node id """
         for r in reversed(self._records):
             if r.node.id == node.id:
                 return r
 
     def get_last_node_record(self, node: p.Node) -> RuntimeRecord:
+        """ Get last record by node id. Raise if no record is found """
         record = self.get_last_node_record_or_none(node)
         if record is None:
-            raise ValueError("Node has no records")
+            raise ValueError(f"Node {node} has no records")
         return record
 
     def get_node_records(self, node: p.Node, test_by_instance=False) -> list[RuntimeRecord]:
@@ -350,22 +353,6 @@ class RuntimeInfo:
                     elif r.has_state(instruction_state):
                         return i
 
-    def patch_node_references(self, program: p.ProgramNode):
-        """ Patch node references to updated program nodes to account for edited method. """
-        logger.info("Patching node references in runtime records")
-        for r in self.records:
-            if r.node is not None and not isinstance(r.node, p.ProgramNode):
-                new_node = program.get_child_by_id(r.node.id)
-                if new_node is None:
-                    logger.error(f"No new node was found to replace {r.node}. Node cannot be patched")
-                else:
-                    if r.node != new_node:
-                        r.node = new_node
-                        logger.debug(f"Patched node reference {new_node}")
-                    else:
-                        logger.warning(f"Node not patched: {new_node} - old node already matched the new node!?")
-        logger.info("Patching complete")
-
     def get_record_by_index(self, index: int) -> RuntimeRecord | None:
         if index < 0:
             raise ValueError(f"index {index} is invalid")
@@ -394,8 +381,27 @@ class RuntimeInfo:
     def print_as_table(self, description: str = ""):
         print(self.get_as_table(description))
 
+    def with_edited_program(self, program: p.ProgramNode) -> RuntimeInfo:
+        """ Return a deep copy with node references set to nodes in program. Does not
+        modify any existing records.
+
+        Note: Uses RuntimeRecordState.clone() which does not change command and command_exec_id.
+        """
+        instance = RuntimeInfo()
+        for r in self.records:
+            # clone record and add the clone
+            record = r.with_edited_node(program)
+            instance._add_record(record, record.exec_id)
+        return instance
+
 class RuntimeRecord:
     def __init__(self, node: p.Node, exec_id: UUID) -> None:
+        """ Create new node. This only happens in
+        - RuntimeInfo.begin_visit() which creates and assigns a random value
+        - null_record() which creates and assigns a random value
+        - clone() / with_edited_node() which copy the existing value
+
+        Note: Changes to node fields must be mirrored in clone() and with_edited_program(). """
         self.exec_id: UUID = exec_id
         self.node = node
         self.name = node.runlog_name
@@ -412,7 +418,7 @@ class RuntimeRecord:
         """ Used for progress for interpreter commands that don't have an command instance """
 
     def __str__(self) -> str:
-        return f'{self.__class__.__name__}(name="{self.name}")'
+        return f'{self.__class__.__name__}(name="{self.name}, exec_id={self.exec_id}")'
 
     @staticmethod
     def null_record() -> RuntimeRecord:
@@ -448,9 +454,6 @@ class RuntimeRecord:
     def add_state_awaiting_condition(self, time: float, tick: int, state_values: TagValueCollection | None):
         self.add_state(RuntimeRecordStateEnum.AwaitingCondition, time, tick, state_values)
 
-    def add_state_awaiting_interrupt(self, time: float, tick: int, state_values: TagValueCollection | None):
-        self.add_state(RuntimeRecordStateEnum.AwaitingInterrrupt, time, tick, state_values)
-
     def add_state_started(self, time: float, tick: int, start_values: TagValueCollection):
         self.add_state(RuntimeRecordStateEnum.Started, time, tick, start_values)
 
@@ -514,6 +517,33 @@ class RuntimeRecord:
         state = self.add_state(RuntimeRecordStateEnum.Failed, time, tick, end_values)
         state.command_exec_id = command_exec_id
 
+    def clone(self) -> RuntimeRecord:
+        instance = RuntimeRecord(self.node, self.exec_id)
+
+        instance.name = self.node.runlog_name
+        instance.visit_start_time = self.visit_start_time
+        instance.visit_start_tick = self.visit_start_tick
+        instance.visit_end_time = self.visit_end_time
+        instance.visit_end_tick = instance.visit_end_tick
+
+        instance.states = [s.clone() for s in self.states]
+        instance.start_values = None if self.start_values is None else self.start_values.clone()
+        instance.end_values = None if self.end_values is None else self.end_values.clone()
+
+        instance.progress = self.progress
+        return instance
+
+
+    def with_edited_node(self, program: p.ProgramNode) -> RuntimeRecord:
+        """ Clones the record, updates the node to the matching node in the new program and returns the clone """
+        instance = self.clone()
+        if self.node.id != "":
+            new_node = program.get_child_by_id(self.node.id)
+            if new_node is None:
+                raise ValueError(f"Failed to copy record for node {self.node} - No new node available")
+            else:
+                instance.node = new_node
+        return instance
 
 class RuntimeRecordState:
     def __init__(
@@ -535,6 +565,15 @@ class RuntimeRecordState:
     def __repr__(self):
         return self.__str__()
 
+    def clone(self) -> RuntimeRecordState:
+        """ Note: command and command_exec_id are copied from existing state. These reference
+        command objects outside or RuntimeInfo. """
+        values = None if self.values is None else self.values.clone()
+        instance = RuntimeRecordState(self.state_name, self.state_time, self.state_tick, values)
+        instance.command = self.command
+        instance.command_exec_id = self.command_exec_id
+        return instance
+
 
 class RuntimeRecordStateEnum(StrEnum):
     """ Defines the states runtime records can take """
@@ -549,8 +588,6 @@ class RuntimeRecordStateEnum(StrEnum):
     """ Waiting for threshold """
     AwaitingCondition = auto()
     """ Waiting for condition """
-    AwaitingInterrrupt = auto()
-    """ Instruction is awaiting invocation by interrupt """
     Started = auto()
     """ Command has started """
     Cancelled = auto()
@@ -622,3 +659,56 @@ class RunLogItemState(StrEnum):
     def has_value(value: str):
         """ Determine if enum has this string value defined. Case sensitive. """
         return hasattr(RunLogItemState, value)
+
+
+def assert_Runlog_HasItem(runtimeinfo: RuntimeInfo, name: str):
+    runlog = runtimeinfo.get_runlog()
+    for item in runlog.items:
+        if item.name == name:
+            return
+    item_names = [item.name for item in runlog.items]
+    raise AssertionError(f"Runlog has no item named '{name}'. It has these names:  {','.join(item_names)}")
+
+def assert_Runlog_HasNoItem(runtimeinfo: RuntimeInfo, name: str):
+    runlog = runtimeinfo.get_runlog()
+    item_names = [item.name for item in runlog.items]
+    for item in runlog.items:
+        if item.name == name:
+            raise AssertionError(f"Runlog has item named '{name}' which was not expected. It has these names:  {','.join(item_names)}")
+
+def assert_Runlog_HasItem_Started(runtimeinfo: RuntimeInfo, name: str):
+    runlog = runtimeinfo.get_runlog()
+    for item in runlog.items:
+        if item.name == name and item.state == RunLogItemState.Started:
+            return
+    raise AssertionError(f"Runlog has no item named '{name}' in Started state")
+
+def assert_Runlog_HasItem_Completed(runtimeinfo: RuntimeInfo, name: str, min_times=1):
+    occurrences = 0
+    runlog = runtimeinfo.get_runlog()
+    for item in runlog.items:
+        if item.name == name and item.state == RunLogItemState.Completed:
+            occurrences += 1
+
+    if occurrences < min_times:
+        raise AssertionError(
+            f"Runlog item named '{name}' did not occur in Completed state at least " +
+            f"{min_times} time(s). It did occur {occurrences} time(s)")
+
+def assert_Runtime_HasRecord(runtimeinfo: RuntimeInfo, name: str):
+    for r in runtimeinfo.records:
+        if r.name == name:
+            return
+    raise AssertionError(f"Runtime has no record named '{name}'")
+
+def assert_Runtime_HasRecord_Started(runtimeinfo: RuntimeInfo, name: str):
+    for r in runtimeinfo.records:
+        if r.name == name and r.has_state(RuntimeRecordStateEnum.Started):
+            return
+    raise AssertionError(f"Runtime has no record named '{name}' in state Started")
+
+def assert_Runtime_HasRecord_Completed(runtimeinfo: RuntimeInfo, name: str):
+    for r in runtimeinfo.records:
+        if r.name == name and r.has_state(RuntimeRecordStateEnum.Started):
+            return
+    raise AssertionError(f"Runtime has no record named '{name}' in state Completed")
