@@ -9,6 +9,7 @@ import time
 
 from openpectus.engine.commands import ContextEngineCommand, CommandArgs
 from openpectus.engine.hardware import HardwareLayerBase, NullHardware, Register, RegisterDirection
+from openpectus.lang.exec.argument_specification import ArgSpec
 from openpectus.lang.exec.base_unit import BaseUnitProvider
 from openpectus.lang.exec.errors import UodValidationError
 from openpectus.lang.exec.readings import (
@@ -471,6 +472,7 @@ class UodCommand(ContextEngineCommand[UnitOperationDefinitionBase]):
         self.finalize_fn: FINAL_FN | None = None
         self.arg_parse_fn: PARSE_FN | None = None
         self._performance_warned = False
+        self.cmd_state: dict[str, Any] = dict()
 
     @staticmethod
     def builder() -> UodCommandBuilder:
@@ -816,6 +818,8 @@ class UodBuilder:
             cb.with_arg_parse_fn(defaultArgumentParser)
         elif arg_parse_fn is not None:
             cb.with_arg_parse_fn(arg_parse_fn)
+        elif arg_parse_fn is None:
+            cb.with_arg_parse_fn(RegexNamedArgumentParser(ArgSpec.NoArgsInstance.regex).parse)
         if cb.name in self.command_factories.keys():
             raise ValueError(f"Duplicate command name: {cb.name}")
         self.command_factories[cb.name] = cb
@@ -895,7 +899,7 @@ class UodBuilder:
             execute_command_name=execute_command_name)
         return self._with_process_value(reading)
 
-    def with_process_value_choice(self, tag_name: str, command_options: dict[str, str] | None = None) -> UodBuilder:
+    def with_process_value_choice(self, tag_name: str, command_options: dict[str, str]) -> UodBuilder:
         """ Add process value for the given tag and enable command choices.
 
         Parameters:
@@ -904,20 +908,13 @@ class UodBuilder:
                 command_options are specified.
             command_options: optional
                 Dictionary that maps command names to their pcode implementation.
-                If not specified, the tag's choices are used as both names and
-                pcode implementation. If specified, the tag's choices are not used and the
-                tag does not need to have choices defined.
         """
         if not self.tags.has(tag_name):
             raise ValueError(f"Cannot add process value choice for tag name {tag_name}. The tag name was not found")
-        tag = self.tags.get(tag_name)
+        if not command_options:
+            raise ValueError(f"Cannot add process value choice without any command_options. Use 'with_process_value' instead.")
         if command_options is not None:
             reading = ReadingWithChoice(tag_name, command_options)
-        elif tag.choices is None or len(tag.choices) == 0:
-            raise ValueError(f"Cannot add process value choice for tag name {tag_name}. " +
-                             "The tag has no choices defined and no options were given")
-        else:
-            reading = ReadingWithChoice(tag_name)
         return self._with_process_value(reading)
 
     def _with_process_value(self, reading: Reading) -> UodBuilder:
@@ -1047,7 +1044,7 @@ class RegexNamedArgumentParser:
         if "option" not in self.get_named_groups():
             return []
         start = self.regex.index("|(") + len("|(")
-        end = self.regex.index("|\\+)+))\\s*")
+        end = self.regex.index(r"|\+)+)(?<!\+))\s*")
         result = unescape(self.regex[start: end]).split("|")
         if "" in result:
             result.remove("")
