@@ -1,12 +1,14 @@
 from __future__ import annotations
+from logging import Logger
 from typing import Self, Type, TypeVar, TypedDict
 
 
 class Position:
-
     def __init__(self, line: int, character: int):
-        self.line: int = line   # todo define 0-1 based. should probably change from before to just match lsp
-        self.character: int = character  # start index = indentation
+        self.line: int = line
+        """ Line number, zero based """
+        self.character: int = character
+        """ Character number on line, zero-based. Corresponds to the indentation of the line """
 
     def is_empty(self) -> bool:
         return self == Position.empty
@@ -105,6 +107,7 @@ class NodeState(TypedDict):
     completed: bool
     cancelled: bool
     forced: bool
+    action_history: list[str]
 
 
 TNode = TypeVar("TNode", bound="Node")
@@ -147,6 +150,16 @@ class SupportCancelForce:
             self._forced = True
 
 
+class SupportsInterrupt():
+    """ Marker interface that indicates that the node type uses the interrupt mechanism.
+
+    Requirements:
+    - it must provide a interrupt_registered property, currently NodeWithChildren does this
+    - interpreter._create_interrupt_handler() must be able to create a handler for the node type
+    """
+    ...
+
+
 class Node(SupportCancelForce):
     instruction_names: list[str] = []
     """ Specifies which node the parser should instantiate for a given instruction name(s) """
@@ -159,10 +172,12 @@ class Node(SupportCancelForce):
         self.position: Position = position
         self.instruction_part: str = ""
         self.instruction_range: Range = Range.empty()
+        self.stripped_instruction_range: Range = Range.empty()
         self.threshold_part: str = ""
         self.arguments_part: str = ""
         self.arguments: str = ""
         self.arguments_range: Range = Range.empty()
+        self.stripped_arguments_range: Range = Range.empty()
         self.comment_part: str = ""
         self.has_comment: bool = False
         self.has_argument: bool = False
@@ -175,9 +190,14 @@ class Node(SupportCancelForce):
         # interpretation state
         self.started: bool = False
         self.completed: bool = False
+        self.action_history: list[str] = []
+        self._empty_node_class_names = [BlankNode.__name__, CommentNode.__name__]
 
     @property
     def name(self) -> str:
+        """ Get node name. Parser guarantees it to not be None and be stripped. May be the empty string"""
+        # parser guards against None
+        assert self.arguments is not None
         return self.arguments
 
     @property
@@ -196,7 +216,9 @@ class Node(SupportCancelForce):
     def has_children(self) -> bool:
         return False
 
-    def get_child_by_id(self, id: str) -> Node | None:
+    def get_child_by_id(self, id: str, include_self=False) -> Node | None:
+        if include_self and self.id == id:
+            return self
         if self.has_children():
             assert isinstance(self, NodeWithChildren)
             for child in self.children:
@@ -207,7 +229,9 @@ class Node(SupportCancelForce):
                     return match
 
     def can_load_state(self, state: NodeState) -> bool:
-        """ Determine whether state is valid for this kind of node """
+        """ Determine whether state is valid for this kind of node """        
+        if state["class_name"] in self._empty_node_class_names:
+            return True
         return state["class_name"] == self.__class__.__name__
 
     def apply_state(self, state: NodeState):
@@ -217,6 +241,7 @@ class Node(SupportCancelForce):
         self.completed = state["completed"]
         self._cancelled = state["cancelled"]
         self._forced = state["forced"]
+        self.action_history = state["action_history"]
 
     def extract_state(self) -> NodeState:
         return NodeState(
@@ -226,7 +251,8 @@ class Node(SupportCancelForce):
             started=self.started,
             completed=self.completed,
             cancelled=self.cancelled,
-            forced=self.forced
+            forced=self.forced,
+            action_history=self.action_history
         )
 
     def get_child_by_instruction(self, instruction_name: str, arguments: str | None = None) -> Node | None:
@@ -255,7 +281,8 @@ class Node(SupportCancelForce):
         self.errors.append(error)
 
     def __str__(self):
-        return f"{self.__class__.__name__}(instruction_name='{self.instruction_name}', arguments={self.arguments}, id='{self.id}')"
+        return f"{self.__class__.__name__}(instruction_name='{self.instruction_name}', " + \
+            f"arguments={self.arguments}, id='{self.id}')"
 
     def __repr__(self):
         return self.__str__()
@@ -273,7 +300,7 @@ class Node(SupportCancelForce):
             parents.append(node.parent)
             node = node.parent
         return parents
-    
+
     @property
     def root(self) -> ProgramNode:
         if isinstance(self, ProgramNode):
@@ -284,28 +311,82 @@ class Node(SupportCancelForce):
 
 
     def reset_runtime_state(self, recursive: bool):
-        # TODO implement, possily by just applying an empty state dict
-        # known state to reset: activated for condition nodes
-        # state = self.extract_state()
-        # state.started = False
-        # state.completed = False
-        # self.apply_state(state)
-
         self._forced = False
 
         self.started = False
         self.completed = False
+        self.action_history = []
 
         if recursive:
             if isinstance(self, NodeWithChildren):
                 for child in self.children:
                     child.reset_runtime_state(True)
 
+    def matches_source(self, other: Node, logger: Logger) -> bool:  # noqa: C901
+        """ Returns True if the two nodes correspond to the same pcode source lines, else False.
+
+        Some white space and comments are not considered changes
+        """
+        logger.debug(f"Comparing nodes {self} and {other}")
+
+        def is_significant_node(node: Node) -> bool:
+            if isinstance(node, WhitespaceNode):
+                return False
+            return True
+
+        def matches(node: Node, other: Node, logger: Logger) -> bool:
+            if node.__class__ != other.__class__:
+                logger.debug(f"Nodes differ by class on line {node.position.line}. " +
+                             f"'{node.__class__}' differs from '{other.__class__}'")
+                return False
+            if node.name != other.name:
+                # Note: This may never happen because if their names differ, they would not be chosen for comparison.
+                logger.debug(f"Nodes differ by name on line {node.position.line}")
+                return False
+            if node.arguments != other.arguments:
+                logger.debug(f"Nodes differ by argument, {node.arguments} differs from {other.arguments} " +
+                             f"on line {node.position.line}")
+                return False
+            if node.threshold != other.threshold:
+                logger.debug(f"Nodes differ by threshold, {node.threshold} differs from {other.threshold} " +
+                             f"on line {node.position.line}")
+                return False
+            if node.position.character != other.position.character:
+                logger.debug(f"Nodes differ by indentation, {node.position.character} differs from {other.position.character} " +
+                             f"on line {node.position.line}")
+            if isinstance(node, NodeWithChildren):
+                assert isinstance(other, NodeWithChildren)  # node and other have the same class
+                node_significant_child_count = len([n for n in node.children if is_significant_node(n)])
+                other_significant_child_count = len([n for n in other.children if is_significant_node(n)])
+                if node_significant_child_count != other_significant_child_count:
+                    logger.debug(f"Nodes differ by number of significant child nodes following line {node.position.line}")
+                    return False
+                match_index = -1
+                for child in node.children:
+                    if not is_significant_node(child):
+                        continue
+                    for index, other_child in enumerate(other.children):
+                        if index <= match_index:
+                            continue
+                        if not is_significant_node(other_child):
+                            continue
+                        match_index = index
+                        result = matches(child, other_child, logger)
+                        if not result:
+                            return False
+                        break
+            return True
+        return matches(self, other, logger)
 
 class NodeWithChildren(Node):
     def __init__(self, position=Position.empty, id=""):
         super().__init__(position, id)
         self._children: list[Node] = []
+        self.interrupt_registered: bool = False
+        """ Whether an interrupt was registered to execute the node. """
+        self.children_complete: bool = False
+        """ Specifies that execution of child nodes should stop or is completed. """
+
         self._last_non_ws_line: int = 0
         """ Populated by WhitespaceAnalyzer """
 
@@ -355,16 +436,33 @@ class NodeWithChildren(Node):
 
     def extract_state(self) -> NodeState:
         state = super().extract_state()
+        state["interrupt_registered"] = self.interrupt_registered  # type: ignore
+        state["children_complete"] = self.children_complete  # type: ignore
         return state
 
     def apply_state(self, state):
+        self.interrupt_registered = state["interrupt_registered"]  # type: ignore
+        self.children_complete = state["children_complete"]  # type: ignore
         super().apply_state(state)
 
+    def reset_runtime_state(self, recursive):
+        self.interrupt_registered = False
+        self.children_complete = False
+        super().reset_runtime_state(recursive)
+
 class ProgramNode(NodeWithChildren):
     def __init__(self, position=Position.empty, id=""):
         super().__init__(position, id)
         self.active_node: Node | None = None
-        """ The node currently executing. Maintained by interpreter. """
+        """ The node currently executing. Is never ProgramNode. Is None untli first instruction is
+        visited. Is not cleared at the end but keeps pointing to the last instruction.
+
+        The value is maintained by the interpreters program iterator. """
+
+        self.revision: int = 0
+        """ The program revision. Starts as 0 and increments every time an edit is performed while running. """
+
+        self.macros: dict[str, MacroNode] = dict()
 
     def get_instructions(self, include_blanks: bool = False) -> list[Node]:
         """ Return list of all program instructions, recursively, depth first. """
@@ -384,13 +482,15 @@ class ProgramNode(NodeWithChildren):
         add_child_nodes(self, nodes)
         return nodes
 
-    def extract_tree_state(self, skip_started_nodes=False) -> dict[str, NodeState]:
-        """ Return map of all nodes keyed by their node id """
+    def extract_tree_state(self) -> dict[str, NodeState]:
+        """ Return map of all nodes' state keyed by their node id.
+
+        This includes non-started nodes that should not be imported.
+        apply_tree_state() handles the filtering on import. """
         result: dict[str, NodeState] = {}
 
         def extract_child_state(node: Node, result: dict[str, NodeState]):
-            if node.started or not skip_started_nodes:
-                result[node.id] = node.extract_state()
+            result[node.id] = node.extract_state()
 
             if isinstance(node, NodeWithChildren):
                 for child in node.children:
@@ -401,14 +501,41 @@ class ProgramNode(NodeWithChildren):
 
     def apply_tree_state(self, state: dict[str, NodeState]):
         def apply_child_state(node: Node):
-            node_state = state.get(node.id, None)
-            if node_state is not None:
-                node.apply_state(node_state)
+            try:
+                node_state = state.get(node.id, None)
+                if node_state is not None:
+                    # Only import state from nodes that have run, i.e nodes before active_node. This leaves
+                    # nodes after active_node alone, which allows changing node types for all nodes that have not started.
+                    if len(node_state["action_history"]) > 0:
+                        node.apply_state(node_state)
+            except KeyError as ke:
+                raise ValueError(f"Failed to apply state {state} to node {node}. Error: {str(ke)}")
             if isinstance(node, NodeWithChildren):
                 for child in node.children:
                     apply_child_state(child)
         apply_child_state(self)
 
+    def reset_runtime_state(self, recursive):
+        self.active_node = None
+        super().reset_runtime_state(recursive)
+
+    def extract_state(self) -> NodeState:
+        state = super().extract_state()
+        state["revision"] = self.revision  # type: ignore
+        state["active_node"] = "" if self.active_node is None else str(self.active_node)  # type: ignore
+        return state
+
+    def apply_state(self, state: NodeState):
+        # Note: while revision is imported from the edited method, method_manager increments it
+        # right after the merge
+        self.revision = int(state["revision"])  # type: ignore
+        # active_node is not imported, it is only used during FFW
+        return super().apply_state(state)
+
+    def __str__(self):
+        return f"{self.__class__.__name__}(instruction_name='{self.instruction_name}', revision={self.revision}, " + \
+            f"id='{self.id}')"
+
     @staticmethod
     def empty() -> ProgramNode:
         return ProgramNode()
@@ -425,6 +552,23 @@ class MarkNode(Node):
 class BlockNode(NodeWithChildren):
     instruction_names = ["Block"]
 
+    def __init__(self, position=Position.empty, id=""):
+        super().__init__(position, id)
+        self.lock_aquired = False
+
+    def reset_runtime_state(self, recursive):
+        self.lock_aquired = False
+        super().reset_runtime_state(recursive)
+
+    def extract_state(self) -> NodeState:
+        state = super().extract_state()
+        state["lock_aquired"] = self.lock_aquired  # type: ignore
+        return state
+
+    def apply_state(self, state: NodeState):
+        self.lock_aquired = bool(state["lock_aquired"])  # type: ignore
+        return super().apply_state(state)
+
 
 class EndBlockNode(Node):
     instruction_names = ["End block"]
@@ -433,29 +577,44 @@ class EndBlockNode(Node):
 class EndBlocksNode(Node):
     instruction_names = ["End blocks"]
 
+
 class BatchNode(Node):
     instruction_names = ["Batch"]
 
 
-class NodeWithCondition(NodeWithChildren):
+class NodeWithTagOperatorValue(Node):
+    operators: list[str]
+
     def __init__(self, position=Position.empty, id=""):
         super().__init__(position, id)
-        self.condition_part: str
-        self.condition: Condition | None
+        self.tag_operator_value_part: str
+        self.tag_operator_value: TagOperatorValue | None
+
+
+class NodeWithCondition(NodeWithTagOperatorValue):
+    operators = ["<=", ">=", "==", "!=", "<", ">", "="]
+
+    def __init__(self, position=Position.empty, id=""):
+        super().__init__(position, id)
+        self.interrupt_registered: bool = False
         self.activated: bool = False
+        """ Node condition was evaluated true"""
 
     def apply_state(self, state: NodeState):
         super().apply_state(state)
         if "activated" not in state.keys():
             raise ValueError(f"Failed to apply state to node {self}. Missing state key 'activated'")
+        self.interrupt_registered = bool(state["interrupt_registered"])  # type: ignore
         self.activated = bool(state["activated"])  # type: ignore
 
     def extract_state(self) -> NodeState:
         state = super().extract_state()
+        state["interrupt_registered"] = self.interrupt_registered  # type: ignore
         state["activated"] = self.activated  # type: ignore
         return state
 
     def reset_runtime_state(self, recursive):
+        self.interrupt_registered = False
         self.activated = False
         self._cancelled = False
         self._forced = False
@@ -470,34 +629,57 @@ class NodeWithCondition(NodeWithChildren):
         return not self.cancelled and not self.forced and not self.activated
 
 
-class WatchNode(NodeWithCondition):
+class NodeWithAssignment(NodeWithTagOperatorValue):
+    operators = ["="]
+
+
+class WatchNode(NodeWithChildren, NodeWithCondition, SupportsInterrupt):
     instruction_names = ["Watch"]
 
 
-class AlarmNode(NodeWithCondition):
+class AlarmNode(NodeWithChildren, NodeWithCondition, SupportsInterrupt):
     instruction_names = ["Alarm"]
 
+    def __init__(self, position=Position.empty, id=""):
+        super().__init__(position, id)
+        self.run_count: int = 0
+        """ The number of times the alarm has completed """
 
-class Condition:
+    def extract_state(self):
+        state = super().extract_state()
+        state["run_count"] = self.run_count  # type: ignore
+        return state
+
+    def apply_state(self, state):
+        self.run_count = int(state["run_count"])
+        super().apply_state(state)
+
+    def reset_runtime_state(self, recursive):
+        # Note: run_count is not reset because it counts alarm invocations
+        super().reset_runtime_state(recursive)
+
+
+class TagOperatorValue:
     def __init__(self):
         self.error = True
         self.lhs = ""
         self.op = ""
         self.rhs = ""
         self.range: Range = Range.empty()
+        self.stripped_range: Range = Range.empty()
 
         self.tag_name: str | None = None
         self.tag_value: str | None = None
         self.tag_unit: str | None = None
         self.tag_value_numeric: int | float | None = None
 
-        # these ranges are problematic. we somehow want to exclude whitespace
-        # for the LSP purpose of underlining lsh/op/rhs but this conflicts
-        # with the calculation of later ones based on the first ones.
-
         self.lhs_range = Range.empty()
+        self.stripped_lhs_range = Range.empty()
         self.op_range = Range.empty()
         self.rhs_range = Range.empty()
+        self.stripped_rhs_range = Range.empty()
+        self.tag_unit_range = Range.empty()
+
 
     def __str__(self):
         return f'{self.__class__.__name__}(lhs="{self.lhs}", op="{self.op}", rhs="{self.rhs}")'
@@ -512,7 +694,8 @@ class WhitespaceNode(Node):
     def __init__(self, position=Position.empty, id=""):
         super().__init__(position, id)
         self.has_only_trailing_whitespace: bool = False
-        """ Specifies that only whitespace instructions follow this whitespace instruction. """
+        """ Specifies that only whitespace instructions follow this whitespace instruction in
+        the current scope as well as outer scopes. """
 
 
 class CommentNode(WhitespaceNode):
@@ -527,7 +710,7 @@ class CommentNode(WhitespaceNode):
         return self
 
 
-class InjectedNode(NodeWithChildren):
+class InjectedNode(NodeWithChildren, SupportsInterrupt):
     pass
 
 
@@ -539,11 +722,81 @@ class MacroNode(NodeWithChildren):
         self.activated: bool = False
         self._cancellable = False
         self._forcible = False
+        self.is_registered: bool = False
+        """ Whether the macro has been registered in the revision. Lifetime is revision. """
+        self.run_started_count: int = 0
+        """ The number of times the macro has started. Life time is the whole run """
 
+    def extract_state(self):
+        state = super().extract_state()
+        state["is_registered"] = self.is_registered  # type: ignore
+        state["run_started_count"] = self.run_started_count  # type: ignore
+        return state
+
+    def apply_state(self, state):
+        self.is_registered = bool(state["is_registered"])
+        self.run_started_count = int(state["run_started_count"])
+        super().apply_state(state)
+
+    def prepare_for_call(self):
+        """ Clears state left over by any previous calls of the macro so it can be called again """
+        self.children_complete = False
+        self.completed = False
+        for macro_child in self.children:
+            macro_child.reset_runtime_state(recursive=True)
+
+    def reset_runtime_state(self, recursive):
+        # Note: is_registered is not reset because that would cause re-registering the macro
+        # Note: run_started_count is not reset because it must maintain the macro invocations count
+        super().reset_runtime_state(recursive)
+
+    def macro_calling_macro(self, macros: dict[str, MacroNode], name: str | None = None) -> list[str]:
+        """ Recurse through macro to produce a path of calls it makes to other macros.
+
+        This is used to identify if a macro will at some point call itself. """
+
+        # Note: This should be done once during analysis and the result be exposed 
+        # and cached on MacroNode/CallMacroNode, possibly as lists of incoming and 
+        # outgoing calls, which could even make this method reduntant
+        # In that case remember injected nodes - should probably rerun analysis on
+        # injection because that may change macro definitions
+
+        name = name if name is not None else self.name
+        assert self.children is not None
+        for child in self.children:
+            if isinstance(child, CallMacroNode):
+                if child.name == name:
+                    return [child.name]
+                elif child.name in macros.keys():
+                    return [child.name] + macros[child.name].macro_calling_macro(macros, name)
+        return []
 
 class CallMacroNode(Node):
     instruction_names = ["Call macro"]
 
+    def __init__(self, position=Position.empty, id=""):
+        super().__init__(position, id)
+        self._cancellable = False
+        self._forcible = False
+        self.activated = False
+
+    def extract_state(self):
+        state = super().extract_state()
+        state["activated"] = self.activated  # type: ignore
+        return state
+
+    def apply_state(self, state):
+        self.activated = bool(state["activated"])  # type: ignore
+        super().apply_state(state)
+
+    def reset_runtime_state(self, recursive):
+        self.activated = False
+        super().reset_runtime_state(recursive)
+
+
+class NotifyNode(Node):
+    instruction_names = ["Notify"]
+
     def __init__(self, position=Position.empty, id=""):
         super().__init__(position, id)
         self._cancellable = False
@@ -558,6 +811,23 @@ class InterpreterCommandNode(CommandBaseNode):
     """ Represents commands that are directly executable by the interpreter. """
     instruction_names = ["Base", "Increment run counter", "Run counter", "Wait"]
 
+    def __init__(self, position=Position.empty, id=""):
+        super().__init__(position, id)
+        self.wait_start_time: float | None = None
+
+    def extract_state(self):
+        state = super().extract_state()
+        state["wait_start_time"] = self.wait_start_time  # type: ignore
+        return state
+
+    def apply_state(self, state):
+        self.wait_start_time = state["wait_start_time"]  # type: ignore
+        super().apply_state(state)
+
+    def reset_runtime_state(self, recursive):
+        self.wait_start_time = None
+        super().reset_runtime_state(recursive)
+
 
 class EngineCommandNode(CommandBaseNode):
     """ Represents internal engine commands that have a command class subclassing InternalEngineCommand. """
@@ -565,6 +835,14 @@ class EngineCommandNode(CommandBaseNode):
                          "Info", "Warning", "Error"]
 
 
+class SimulateNode(NodeWithAssignment):
+    instruction_names = ["Simulate"]
+
+
+class SimulateOffNode(Node):
+    instruction_names = ["Simulate off"]
+
+
 class UodCommandNode(CommandBaseNode):
     """ Represents a uod command, subclassing UodCommand. """
 
