@@ -21,6 +21,7 @@ from openpectus.lsp.model import (
     CodeAction, WorkspaceEdit,
     get_item_range, get_item_severity
 )
+from openpectus.lsp.model import Command as LSPCommand
 import openpectus.lang.model.ast as p
 import openpectus.protocol.models as ProMdl
 import openpectus.aggregator.deps as agg_deps
@@ -58,9 +59,7 @@ def fetch_process_value(engine_id: str, tag_name) -> ProMdl.TagValue | None:
     if not engine_data:
         return None
 
-    for tag_name_, tag_value in engine_data.tags_info.map.items():
-        if tag_name_ == tag_name:
-            return tag_value
+    return engine_data.tags_info.map.get(tag_name, None)
 
 
 def build_tags(uod_def: ProMdl.UodDefinition) -> TagValueCollection:
@@ -321,13 +320,13 @@ def hover(document: Document, position: Position, engine_id: str) -> Hover | Non
                     kind="markdown",
                     value="```pcode\r\n"+docstring+"\r\n```" if docstring else "",
                 ),
-                range=lsp_range_from_ast_range(node.instruction_range),
+                range=lsp_range_from_ast_range(node.stripped_instruction_range),
             )
         # Hovering condition
-        if isinstance(node, p.NodeWithCondition) and node.condition:
+        if isinstance(node, p.NodeWithTagOperatorValue) and node.tag_operator_value:
             # Show current tag value
-            if node.condition.lhs and analysis_input.tags.has(node.condition.lhs) and position_ast in node.condition.lhs_range:
-                process_value = fetch_process_value(engine_id, node.condition.lhs)
+            if node.tag_operator_value.tag_name and analysis_input.tags.has(node.tag_operator_value.tag_name) and position_ast in node.tag_operator_value.stripped_lhs_range:
+                process_value = fetch_process_value(engine_id, node.tag_operator_value.tag_name)
                 if not process_value:
                     return
                 value_str = ""
@@ -339,16 +338,16 @@ def hover(document: Document, position: Position, engine_id: str) -> Hover | Non
                     # If the value specified on the RHS has a different measurement unit
                     # than the tag unit, then we want to show the unit in its own unit
                     # as well as the one the user want to compare to.
-                    if (node.condition.tag_unit and
-                       process_value.value_unit != node.condition.tag_unit and
-                       node.condition.tag_unit in units_compaible_with_tag(analysis_input, node.condition.lhs) and
+                    if (node.tag_operator_value.tag_unit and
+                       process_value.value_unit != node.tag_operator_value.tag_unit and
+                       node.tag_operator_value.tag_unit in units_compaible_with_tag(analysis_input, node.tag_operator_value.lhs) and
                        isinstance(process_value.value, (int, float))):
                         converted_value = convert_value_to_unit(
                             process_value.value,
                             process_value.value_unit,
-                            node.condition.tag_unit
+                            node.tag_operator_value.tag_unit
                         )
-                        value_str += " (" + f"{converted_value:0.2f}".replace(".", ",") + f" {node.condition.tag_unit})"
+                        value_str += " (" + f"{converted_value:0.2f}".replace(".", ",") + f" {node.tag_operator_value.tag_unit})"
                 elif process_value.value:
                     value_str = str(process_value.value)
                 return Hover(
@@ -356,28 +355,28 @@ def hover(document: Document, position: Position, engine_id: str) -> Hover | Non
                         kind="markdown",
                         value=f"Current value: {value_str}",
                     ),
-                    range=lsp_range_from_ast_range(node.condition.lhs_range),
+                    range=lsp_range_from_ast_range(node.tag_operator_value.stripped_lhs_range),
                 )
             # Show text desciption of comparison operator
-            if position_ast in node.condition.op_range:
+            if position_ast in node.tag_operator_value.op_range:
                 return Hover(
                     contents=MarkupContent(
                         kind="plaintext",
-                        value=operator_descriptions[node.condition.op],
+                        value=operator_descriptions[node.tag_operator_value.op],
                     ),
-                    range=lsp_range_from_ast_range(node.condition.op_range),
+                    range=lsp_range_from_ast_range(node.tag_operator_value.op_range),
                 )
             # Show compatible units of measurement
-            unit_options = units_compaible_with_tag(analysis_input, node.condition.lhs)
+            unit_options = units_compaible_with_tag(analysis_input, node.tag_operator_value.lhs)
             unit_options_str = ", ".join(unit_options)
-            if unit_options_str and position_ast in node.condition.rhs_range:
+            if unit_options_str and position_ast in node.tag_operator_value.stripped_rhs_range:
                 if len(unit_options) >= 1:
                     return Hover(
                         contents=MarkupContent(
                             kind="markdown",
                             value=f"Unit{'s' if len(unit_options) > 1 else ''}: {unit_options_str}.",
                         ),
-                        range=lsp_range_from_ast_range(node.condition.rhs_range),
+                        range=lsp_range_from_ast_range(node.tag_operator_value.stripped_rhs_range),
                     )
         # Hovering "Call macro"
         elif isinstance(node, p.CallMacroNode) and position_ast in node.arguments_range:
@@ -401,7 +400,7 @@ def hover(document: Document, position: Position, engine_id: str) -> Hover | Non
                             kind="markdown",
                             value=f"Specify a value with unit '{units_str}'.",
                         ),
-                        range=lsp_range_from_ast_range(node.arguments_range),
+                        range=lsp_range_from_ast_range(node.stripped_arguments_range),
                     )
                 elif len(units) > 1:
                     return Hover(
@@ -409,7 +408,7 @@ def hover(document: Document, position: Position, engine_id: str) -> Hover | Non
                             kind="markdown",
                             value=f"Specify a value with one of the following units: {units_str}.",
                         ),
-                        range=lsp_range_from_ast_range(node.arguments_range),
+                        range=lsp_range_from_ast_range(node.stripped_arguments_range),
                     )
                 additive_options = arg_parser.get_additive_options()
                 exclusive_options = arg_parser.get_exclusive_options()
@@ -420,7 +419,7 @@ def hover(document: Document, position: Position, engine_id: str) -> Hover | Non
                             kind="markdown",
                             value=f"Specify one or more (separate with +) of the following options: {options_str}.",
                         ),
-                        range=lsp_range_from_ast_range(node.arguments_range),
+                        range=lsp_range_from_ast_range(node.stripped_arguments_range),
                     )
                 elif not additive_options and exclusive_options:
                     options_str = ", ".join(exclusive_options)
@@ -429,7 +428,7 @@ def hover(document: Document, position: Position, engine_id: str) -> Hover | Non
                             kind="markdown",
                             value=f"Specify one of the following options: {options_str}",
                         ),
-                        range=lsp_range_from_ast_range(node.arguments_range),
+                        range=lsp_range_from_ast_range(node.stripped_arguments_range),
                     )
                 elif additive_options and exclusive_options:
                     options_str = ", ".join(additive_options+exclusive_options)
@@ -438,7 +437,7 @@ def hover(document: Document, position: Position, engine_id: str) -> Hover | Non
                             kind="markdown",
                             value=f"Specify one or possibly more of the following options: {options_str}.",
                         ),
-                        range=lsp_range_from_ast_range(node.arguments_range),
+                        range=lsp_range_from_ast_range(node.stripped_arguments_range),
                     )
 
 
@@ -449,12 +448,17 @@ def completions(document: Document, position: Position, ignored_names, engine_id
     if line is None:
         # Show all possible commands
         return [
-            CompletionItem(label=command_name, kind=CompletionItemKind.Function, preselect=False)
+            CompletionItem(
+                label=command_name,
+                insertText=command_name+": " if analysis_input.commands.get(command_name).accepts_arguments else command_name,
+                kind=CompletionItemKind.Function,
+                command=LSPCommand(title="", command="editor.action.triggerSuggest"),
+            )
             for command_name in analysis_input.commands.names
         ]
 
     # get whole query, eg "St" or "Alarm: Run T"
-    char: int = position["character"]
+    char = position["character"]
     if char < len(line):
         query = line[0:char]
     else:
@@ -462,22 +466,25 @@ def completions(document: Document, position: Position, ignored_names, engine_id
     pcode_parser = PcodeParser()
     node = pcode_parser._parse_line(line, position["line"])
     position_ast = ast_position_from_lsp_position(position)
+    leading_space = position["character"] > 0 and line[position["character"]-1:position["character"]] == " "
+    leading_plus = position["character"] > 0 and line[position["character"]-1:position["character"]] == "+"
 
     if node:
         if node.instruction_name in analysis_input.commands.names:
             # Completion of Watch/Alarm which are special because of conditions
-            if isinstance(node, p.NodeWithCondition) and node.condition:
+            if isinstance(node, p.NodeWithTagOperatorValue) and node.tag_operator_value:
                 # Complete tag name
-                if position_ast in node.condition.lhs_range or (node.condition.lhs == "" and node.arguments_part.strip() not in analysis_input.tags.names):
-                    prefix = " " if query.endswith(":") else ""
-                    if node.condition.lhs_range.is_empty():
+                if position_ast in node.tag_operator_value.stripped_lhs_range or (node.tag_operator_value.lhs == "" and node.arguments_part.strip() not in analysis_input.tags.names):
+                    prefix = "" if leading_space else " "
+                    if node.tag_operator_value.lhs_range.is_empty():
                         return [
                             CompletionItem(
                                 label=name,
-                                insertText=prefix+name,
+                                insertText=prefix+name+" ",
                                 kind=CompletionItemKind.Enum,
+                                command=LSPCommand(title="", command="editor.action.triggerSuggest") if node.tag_operator_value.op == "" else None
                             )
-                            for name in analysis_input.get_tag_completions(node.condition.lhs)
+                            for name in analysis_input.get_tag_completions(node.tag_operator_value.lhs)
                         ]
                     else:
                         return [
@@ -485,41 +492,44 @@ def completions(document: Document, position: Position, ignored_names, engine_id
                                 label=name,
                                 kind=CompletionItemKind.Enum,
                                 textEdit=TextEdit(
-                                    range=lsp_range_from_ast_range(node.condition.lhs_range),
-                                    newText=prefix+name
+                                    range=lsp_range_from_ast_range(node.tag_operator_value.lhs_range),
+                                    newText=" "+name+" "
                                 ),
+                                command=LSPCommand(title="", command="editor.action.triggerSuggest") if node.tag_operator_value.op == "" else None
                             )
-                            for name in analysis_input.get_tag_completions(node.condition.lhs)
+                            for name in analysis_input.get_tag_completions(node.tag_operator_value.lhs)
                         ]
                 # Complete operator
-                elif position_ast in node.condition.op_range or node.condition.op_range.is_empty():
-                    prefix = "" if query.endswith(" ") else " "
-                    if node.condition.op_range.is_empty():
+                elif position_ast in node.tag_operator_value.op_range or node.tag_operator_value.op == "":
+                    prefix = "" if leading_space else " "
+                    if node.tag_operator_value.op_range.is_empty():
                         return [
                             CompletionItem(
-                                label=f"{operator} ({operator_description})",
-                                insertText=prefix+operator,
+                                label=f"{operator} ({operator_descriptions[operator]})",
+                                insertText=prefix+operator+" ",
                                 kind=CompletionItemKind.Enum,
+                                command=LSPCommand(title="", command="editor.action.triggerSuggest") if node.tag_operator_value.tag_unit is None else None
                             )
-                            for operator, operator_description in operator_descriptions.items()
+                            for operator in node.operators
                         ]
                     else:
                         return [
                             CompletionItem(
-                                label=f"{operator} ({operator_description})",
+                                label=f"{operator} ({operator_descriptions[operator]})",
                                 kind=CompletionItemKind.Enum,
                                 textEdit=TextEdit(
-                                    range=lsp_range_from_ast_range(node.condition.op_range),
-                                    newText=prefix+operator
+                                    range=lsp_range_from_ast_range(node.tag_operator_value.op_range),
+                                    newText=operator
                                 ),
+                                command=LSPCommand(title="", command="editor.action.triggerSuggest") if node.tag_operator_value.tag_unit is None else None
                             )
-                            for operator, operator_description in operator_descriptions.items()
+                            for operator in node.operators
                         ]
                 # Complete unit
-                elif position_ast in node.condition.rhs_range or position_ast > node.condition.op_range:
-                    prefix = "" if query.endswith(" ") else " "
-                    unit_options = units_compaible_with_tag(analysis_input, node.condition.lhs)
-                    if len(unit_options) > 0:
+                elif position_ast in node.tag_operator_value.rhs_range or position_ast > node.tag_operator_value.op_range:
+                    prefix = "" if leading_space else " "
+                    unit_options = units_compaible_with_tag(analysis_input, node.tag_operator_value.lhs)
+                    if len(unit_options) > 0 and node.tag_operator_value.tag_unit is None:
                         return [
                             CompletionItem(
                                 label=unit,
@@ -544,7 +554,7 @@ def completions(document: Document, position: Position, ignored_names, engine_id
                 macro_visitor = MacroVisitor()
                 for _ in macro_visitor.visit(program):
                     pass
-                prefix = " " if query.endswith(":") else ""
+                prefix = "" if leading_space else " "
                 if node.arguments_range.is_empty():
                     return [
                         CompletionItem(
@@ -559,28 +569,57 @@ def completions(document: Document, position: Position, ignored_names, engine_id
                         CompletionItem(
                             label=name,
                             kind=CompletionItemKind.Enum,
-                            textEdit=TextEdit(range=lsp_range_from_ast_range(node.arguments_range), newText=prefix+name),
+                            textEdit=TextEdit(
+                                range=lsp_range_from_ast_range(node.arguments_range),
+                                newText=prefix+name
+                            ),
                         )
                         for name in list(macro_visitor.macros.keys())
                     ]
-            # Completion of all other commands
-            elif node.instruction_name in analysis_input.commands.names:
-                arg_parser = analysis_input.commands.get(node.instruction_name).arg_parser
-                prefix = " " if query.endswith(":") else ""
-                if arg_parser:
-                    options = arg_parser.get_additive_options()+arg_parser.get_exclusive_options()+arg_parser.get_units()
-                    # Complete additive options
-                    if node.arguments_part.endswith("+"):
+            elif isinstance(node, p.SimulateOffNode):
+                # Complete tag name
+                if position_ast in node.arguments_range or node.arguments == "" or node.arguments not in analysis_input.tags.names:
+                    prefix = "" if leading_space else " "
+                    if node.arguments_range.is_empty():
                         return [
                             CompletionItem(
                                 label=name,
                                 insertText=prefix+name,
                                 kind=CompletionItemKind.Enum,
                             )
-                            for name in arg_parser.get_additive_options()
+                            for name in analysis_input.get_tag_completions(node.arguments)
+                        ]
+                    else:
+                        return [
+                            CompletionItem(
+                                label=name,
+                                kind=CompletionItemKind.Enum,
+                                textEdit=TextEdit(
+                                    range=lsp_range_from_ast_range(node.arguments_range),
+                                    newText=prefix+name
+                                ),
+                            )
+                            for name in analysis_input.get_tag_completions(node.arguments)
+                        ]
+
+            # Completion of all other commands
+            elif node.instruction_name in analysis_input.commands.names:
+                arg_parser = analysis_input.commands.get(node.instruction_name).arg_parser
+                prefix = "" if leading_space or leading_plus else " "
+                if arg_parser:
+                    options = arg_parser.get_additive_options()+arg_parser.get_exclusive_options()+arg_parser.get_units()
+                    # Complete additive options
+                    if node.arguments.endswith("+"):
+                        return [
+                            CompletionItem(
+                                label=name,
+                                insertText=prefix+name,
+                                kind=CompletionItemKind.Enum,
+                            )
+                            for name in arg_parser.get_additive_options() if name not in node.arguments
                         ]
                     # Complete additive, exclusive and units
-                    if not contains_any(node.arguments_part.strip(), options):
+                    if not contains_any(node.arguments, options):
                         return [
                             CompletionItem(
                                 label=name,
@@ -589,22 +628,28 @@ def completions(document: Document, position: Position, ignored_names, engine_id
                             )
                             for name in options
                         ]
-        elif node.instruction_name and node.arguments_part == "":
+        elif node.instruction_name and node.arguments == "":
             # Completion of command name
             return [
                 CompletionItem(
-                    label=word,
+                    label=command_name,
                     kind=CompletionItemKind.Function,
-                    textEdit=TextEdit(range=lsp_range_from_ast_range(node.instruction_range), newText=word),
+                    textEdit=TextEdit(
+                        range=lsp_range_from_ast_range(node.instruction_range),
+                        newText=command_name+": " if analysis_input.commands.get(command_name).accepts_arguments else command_name,
+                    ),
+                    command=LSPCommand(title="", command="editor.action.triggerSuggest"),
                 )
-                for word in analysis_input.get_command_completions(node.instruction_name)
+                for command_name in analysis_input.get_command_completions(node.instruction_name)
                 ]
     if query.strip() == "":
         # Blank line. Show all possible commands
         return [
             CompletionItem(
                 label=command_name,
+                insertText=command_name+": " if analysis_input.commands.get(command_name).accepts_arguments else command_name,
                 kind=CompletionItemKind.Function,
+                command=LSPCommand(title="", command="editor.action.triggerSuggest")
             )
             for command_name in analysis_input.commands.names
         ]
