@@ -27,6 +27,9 @@ from openpectus.test.engine.utility_methods import (
 )
 from typing_extensions import override
 
+# the number of ticks to run befpre the first command (not including a blank first line)
+start_ticks = 1
+
 
 configure_test_logger()
 set_engine_debug_logging()
@@ -94,10 +97,16 @@ def create_test_uod() -> UnitOperationDefinitionBase:  # noqa C901
             from_tag=lambda x: 1 if x == "Reset" else 0,
             to_tag=lambda x: "Reset" if x == 1 else "N/A",
         )
+        .with_hardware_register(
+            "Danger",
+            RegisterDirection.Write,
+            path="Objects;2:System;2:Danger",
+            safe_value=False,
+        )
         # Readings
         .with_tag(ReadingTag("FT01", "L/h"))
         .with_tag(SelectTag("Reset", value="N/A", unit=None, choices=["Reset", "N/A"]))
-        .with_tag(Tag("Danger", value=True, unit=None, direction=TagDirection.Output, safe_value=False))
+        .with_tag(Tag("Danger", value=True, unit=None, direction=TagDirection.Output))
         .with_command(name="Reset", exec_fn=reset)
         .with_command(name="CmdWithArgs", exec_fn=cmd_with_args, arg_parse_fn=cmd_arg_parse)
         .with_command(name="overlap1", exec_fn=overlap_exec)
@@ -208,7 +217,7 @@ class TestEngine(unittest.TestCase):
 
         self.assertEqual("N/A", reset_tag.get_value())
 
-        run_engine(e, "Reset", 3)
+        run_engine(e, "Reset", start_ticks + 1)
         self.assertEqual("Reset", reset_tag.get_value())
 
         # assert tags marked dirty
@@ -245,7 +254,7 @@ class TestEngine(unittest.TestCase):
     def test_write_process_values_writes_data_to_registers(self):
         e = self.engine
 
-        run_engine(e, "", 1)
+        run_engine(e, "", start_ticks)
         time.sleep(0.1)
 
         # set hw values
@@ -287,6 +296,16 @@ class TestEngine(unittest.TestCase):
 
         self.assertEqual(e.tags[SystemTagName.METHOD_STATUS].get_value(), MethodStatusEnum.OK)
 
+    def test_uod_command_method_state(self):
+        e = self.engine
+        program = "CmdWithArgs: a, b,c ,  d"
+        run_engine(e, program, 10)
+
+        method_state = e.method_manager.get_method_state()
+        node = e.method_manager.program.get_child_by_instruction("CmdWithArgs")
+        assert node is not None
+        self.assertTrue(node.id in method_state.executed_line_ids)
+
     def test_uod_command_w_arguments_fail(self):
         e = self.engine
         program = "CmdWithArgs: FAIL"
@@ -480,7 +499,7 @@ class TestEngine(unittest.TestCase):
 
         self.assertEqual("N/A", e.uod.tags["Reset"].get_value())
 
-        run_engine(e, "Reset", 2)
+        run_engine(e, "Reset", start_ticks + 1)
         self.assertEqual("Reset", e.uod.tags["Reset"].get_value())
 
         # Reset takes 5 ticks to revert
@@ -507,13 +526,13 @@ Reset
 
         print_runtime_records(e)
 
-        r = records[2]
+        r = records[1]
         self.assertEqual("Reset", r.name)
         self.assertTrue(r.has_state(RuntimeRecordStateEnum.Started))
         self.assertTrue(r.has_state(RuntimeRecordStateEnum.Cancelled))
         self.assertTrue(not r.has_state(RuntimeRecordStateEnum.Completed))
 
-        r = records[3]
+        r = records[2]
         self.assertEqual("Reset", r.name)
         self.assertTrue(r.has_state(RuntimeRecordStateEnum.Started))
         self.assertTrue(not r.has_state(RuntimeRecordStateEnum.Cancelled))
@@ -534,7 +553,7 @@ overlap2
 
         print_runtime_records(e)
 
-        r = rs[2]
+        r = rs[1]
         self.assertEqual("overlap1", r.name)
         self.assertTrue(r.has_state(RuntimeRecordStateEnum.Started))
         self.assertTrue(r.has_state(RuntimeRecordStateEnum.Cancelled))
@@ -544,7 +563,7 @@ overlap2
         continue_engine(e, 1)
         print_runtime_records(e)
 
-        r = rs[3]
+        r = rs[2]
         self.assertEqual("overlap2", r.name)
         self.assertTrue(r.has_state(RuntimeRecordStateEnum.Started))
         self.assertTrue(not r.has_state(RuntimeRecordStateEnum.Cancelled))
@@ -571,14 +590,14 @@ Mark: X
 
         print_runtime_records(e)
 
-        self.assertEqual("Block: A", rs[2].name)
+        self.assertEqual("Block: A", rs[1].name)
         self.assertFalse(rs[0].has_state(RuntimeRecordStateEnum.Started))
 
         continue_engine(e, 1)
-        self.assertTrue(rs[2].has_state(RuntimeRecordStateEnum.Started))
+        self.assertTrue(rs[1].has_state(RuntimeRecordStateEnum.Started))
 
-        continue_engine(e, 1)
-        self.assertTrue(rs[2].has_state(RuntimeRecordStateEnum.Completed))
+        continue_engine(e, 2)
+        self.assertTrue(rs[1].has_state(RuntimeRecordStateEnum.Completed))
 
         print_runtime_records(e)
 
@@ -593,9 +612,8 @@ Mark: X
 
         print_runtime_records(e)
 
-        r = e.interpreter.runtimeinfo.records[2]
+        r = e.interpreter.runtimeinfo.records[1]
         self.assertEqual("Watch: Block Time > 0.2s", r.name)
-        self.assertTrue(r.has_state(RuntimeRecordStateEnum.AwaitingInterrrupt))
         self.assertTrue(r.has_state(RuntimeRecordStateEnum.AwaitingCondition))
         # self.assertFalse(r.has_state(RuntimeRecordStateEnum.Started))
 
@@ -638,6 +656,7 @@ Reset
 
         e.inject_code("Increment run counter")
         e.tick(0, 0)
+        e.tick(0, 0)
 
         self.assertEqual(1, e._system_tags["Run Counter"].get_value())
 
@@ -664,7 +683,8 @@ Mark: C
     def test_set_and_continue_program(self):
         code = "Mark: A\nMark: B\nMark: C"
         e = self.engine
-        run_engine(e, code, 3)
+        run_engine(e, code, 2)
+        # run_engine(e, code, start_ticks)
         self.assertEqual(['A'], e.interpreter.get_marks())
         continue_engine(e, 10)
 
@@ -897,6 +917,7 @@ Wait: .5s
 Mark: b
 """
         run_engine(e, program, 2)
+        #run_engine(e, program, start_ticks)
         self.assertEqual([], e.interpreter.get_marks())
 
         continue_engine(e, 1)
@@ -930,20 +951,6 @@ Mark: b
 
         self.assertFalse(danger_tag.get_value())
 
-    def test_restart_can_stop(self):
-        set_engine_debug_logging()
-        set_interpreter_debug_logging()
-        program = """
-Mark: A
-Restart
-"""
-        e = self.engine
-        run_engine(e, program, 3)
-
-        # when no commands need to be stopped, restart immediately moves to Stopped
-        system_state = e.tags[SystemTagName.SYSTEM_STATE]
-        self.assertEqual(system_state.get_value(), SystemStateEnum.Restarting)
-
 
     # --- Totalizers ---
 
@@ -1068,7 +1075,7 @@ Base: s
                .with_accumulated_volume(totalizer_tag_name="calc")
                .build())
 
-        program = """
+        program = """\
 Base: s
 Wait: 0.45s
 Block: A
@@ -1098,7 +1105,6 @@ Block: A
             self.assertAlmostEqual(acc_vol.as_float(), 0.7, delta=delta)
             self.assertAlmostEqual(block_vol.as_float(), 0.7, delta=delta)
 
-
             continue_engine(e, 1)  # Block
             self.assertEqual(block.get_value(), "A")
             self.assertAlmostEqual(acc_vol.as_float(), 0.8, delta=delta)
@@ -1204,7 +1210,7 @@ Base: CV
                .with_accumulated_cv(cv_tag_name="CV", totalizer_tag_name="calc")
                .build())
 
-        program = """
+        program = """\
 Base: s
 Block: A
     0.45 End block
@@ -1271,80 +1277,12 @@ Block: A
                 percent_tag = uod.tags["X3"]
                 self.assertEqual(percent_tag.unit, "%")
 
-
-    # --- Restart ---
-
-
-    def test_restart_can_restart(self):
-        set_engine_debug_logging()
-        set_interpreter_debug_logging()
-        program = """
-Mark: A
-Increment run counter
-Restart
-"""
-        e = self.engine
-        self.assertEqual(0, e.tags[SystemTagName.RUN_COUNTER].as_number())
-
-        run_engine(e, program, 4)
-
-        run_id_1 = e.tags[SystemTagName.RUN_ID].get_value()
-        self.assertEqual(e.tags[SystemTagName.SYSTEM_STATE].get_value(), SystemStateEnum.Restarting)
-        self.assertEqual(1, e.tags[SystemTagName.RUN_COUNTER].as_number())
-
-        self.assertEqual(e.interpreter.get_marks(), ["A"])
-
-        continue_engine(e, 1)
-
-        self.assertEqual(e.tags[SystemTagName.SYSTEM_STATE].get_value(), SystemStateEnum.Stopped)
-        self.assertIsNone(e.tags[SystemTagName.RUN_ID].get_value())
-        self.assertEqual(e.interpreter.get_marks(), [])
-        self.assertEqual(1, e.tags[SystemTagName.RUN_COUNTER].as_number())
-
-        continue_engine(e, 1)
-
-        run_id2 = e.tags[SystemTagName.RUN_ID].get_value()
-        self.assertEqual(e.tags[SystemTagName.SYSTEM_STATE].get_value(), SystemStateEnum.Running)
-        self.assertNotEqual(run_id_1, run_id2)
-        self.assertEqual(e.interpreter.get_marks(), [])
-        self.assertEqual(1, e.tags[SystemTagName.RUN_COUNTER].as_number())
-
-        continue_engine(e, 3)
-
-        self.assertEqual(e.interpreter.get_marks(), ["A"])
-        self.assertEqual(2, e.tags[SystemTagName.RUN_COUNTER].as_number())
-
-    def test_restart_stop_ticking_interpreter(self):
-
-        set_interpreter_debug_logging()
-        program = """
-Mark: A
-Restart
-Mark: X
-"""
-        e = self.engine
-        run_engine(e, program, 1)
-
-        for _ in range(30):
-            continue_engine(e, 1)
-            self.assertTrue("X" not in e.interpreter.get_marks())
-
-    def test_restart_cancels_running_commands(self):
-        program = """
-Reset
-Restart
-"""
-        e = self.engine
-        run_engine(e, program, 3)
-
-        system_state = e.tags[SystemTagName.SYSTEM_STATE]
-        self.assertEqual(system_state.get_value(), SystemStateEnum.Restarting)
-
     def test_EngineCommandEnum_has(self):
         self.assertTrue(EngineCommandEnum.has_value("Stop"))
         self.assertFalse(EngineCommandEnum.has_value("stop"))
         self.assertFalse(EngineCommandEnum.has_value("STOP"))
 
+
 # ----------- Engine Error -------------
 
     def test_engine_error_causes_Paused_state(self):
