@@ -44,12 +44,15 @@ InstructionName = Literal[
     "Stop", "Restart",
     "Info", "Warning", "Error",
     "Macro", "Call macro",
+    "Increment run counter",
+    "Noop", "Notify",
+    "Simulate", "Simulate off"
 ]
 """ Defines the awaitable instructions of the test engine runner """
 
 FindInstructionState = Literal[
     "any", "started", "completed", "failed", "cancelled",
-    "awaiting_threshold", "awaiting_condition", "awaiting_interrupt",
+    "awaiting_threshold", "awaiting_condition",
 ]
 """ Defines the awaitable instruction states of the test engine runner """
 
@@ -90,10 +93,12 @@ class EngineTestInstance(EventListener):
         self.engine.set_method(method)
 
         self._search_index = 0
-        self.engine.emitter.add_listener(self)  # register as listener for lifetime events, so they can be awaited
+        # register as listener for lifetime events, so these events can be awaited in the tests
+        self.engine.emitter.add_listener(self)
         self._last_event: EventName | None = None
 
     def start(self):
+        """ Schedules the Start command. run_* is required to actually run it """
         self.engine.schedule_execution(EngineCommandEnum.START)
         self.timing.timer.start()
 
@@ -146,8 +151,10 @@ class EngineTestInstance(EventListener):
             if self.engine.has_error_state():
                 ex = self.engine.get_error_state_exception()
                 if ex is None:
+                    logger.error("Engine failed with an unspecified error", exc_info=True)
                     raise EngineError("Engine failed with an unspecified error")
                 else:
+                    logger.error(f"Engine failed with exception: {ex}", exc_info=True)
                     raise EngineError(f"Engine failed with exception: {ex}", exception=ex)
 
         return ticks
@@ -206,14 +213,15 @@ class EngineTestInstance(EventListener):
                     self._search_index = 0
                 return True
 
-        logger.debug(f"Start waiting for instruction {instruction_name}, state: {state}, arguments: {arguments}")
+        logger.info(f"Start waiting for instruction {instruction_name}, state: {state}, arguments: {arguments}")
         try:
             ticks = self.run_until_condition(cond, max_ticks=max_ticks)
         except TimeoutError:
             raise TimeoutError(
                 f"Timeout while waiting for instruction '{instruction_name}', state: {state}, arguments: {arguments}")
 
-        logger.debug(f"Done waiting for instruction {instruction_name}, state: {state}, arguments: {arguments}")
+        logger.info(f"Done waiting for instruction {instruction_name}, state: {state}, arguments: {arguments}. " +
+                    f"Duration: {ticks} ticks.")
         return ticks
 
     def index_step_back(self, steps=1):
@@ -254,17 +262,21 @@ class EngineTestInstance(EventListener):
         """ Restart the method and wait a few ticks for engine to come back up. """
         self.engine.schedule_execution(EngineCommandEnum.RESTART)
         self._search_index = 0
-        self.run_ticks(3)
+        self._last_event = None
+        self.run_ticks(1)
+        while self._last_event is None or self._last_event != "start":
+            print("Waiting for started event")
+            self.run_ticks(1)
 
     def run_ticks(self, ticks: int) -> None:
         """ Continue program execution until te specified number of ticks. """
 
-        logger.debug(f"Start waiting for {ticks} ticks")
+        logger.info(f"Start waiting for {ticks} ticks")
         max_ticks = ticks + 1
         try:
             _ = self.run_until_condition(lambda: False, max_ticks)
         except TimeoutError:
-            logger.debug(f"Done waiting for {ticks} ticks")
+            logger.info(f"Done waiting for {ticks} ticks")
             return
 
         raise ValueError("Could not wait??")
